const TAG = "listener";

let censored_cache = new Map();
let context_requests = [];
let do_cache = true;
let current_tab_id = null;

/**********************************************************************************************************************
 * OPTIONS
 **********************************************************************************************************************/
browser.runtime.onMessage.addListener(handleMessage);

function handleMessage(request, sender, sendResponse) {
    if (request.getdata) {
        sendData(sendResponse);
        return true;
    }
    if (request.clearcache) {
        censored_cache = new Map();
    }
    if (request.options) {
        updateOptions(false);
    }
    if (request.logout) {
        logout();
    }
    if (request.login) {
        login(request.username, request.password, sendResponse);
        return true;
    }
}

function sendData(sendResponse) {
    sendResponse({
        cache: censored_cache,
        queue: queue
    });
}

let extension_active = true;
let labels = null;
let censor_type = null;
let file_types = null;


function initializeOptions() {
    let options_request = browser.storage.sync.get(['active', 'labels', 'censor_type', 'file_types', 'debug', 'do_cache',
        'bar_configuration', 'blur_configuration', 'pixel_configuration', 'glitch_configuration', 'gif_configuration', 'whiteblacklist_configuration']);
    options_request.then((res) => {

        // Debugging
        if (res.debug != null) {
            debug = res.debug;
        } else {
            debug = false;
            browser.storage.sync.set({
                debug: false,
            });
        }

        //Cache
        if (res.do_cache != null) {
            do_cache = res.do_cache;
        } else {
            do_cache = true;
            browser.storage.sync.set({
                do_cache: true,
            });
        }

        // Active
        if (res.active != null) {
            extension_active = res.active;
        } else {
            extension_active = true;
            browser.storage.sync.set({
                active: true,
            });
        }
        // Labels
        if (res.labels != null) {
            labels = res.labels;
        } else {
        	// TODO obtain from list of areas
            labels = ["BELLYEXPOSED",
                "BELLYCOVERED",
                "BUTTOCKSEXPOSED",
                "BUTTOCKSCOVERED",
                "FEMALEBREASTEXPOSED",
                "FEMALEBREASTCOVERED",
                "FEMALEGENITALIAEXPOSED",
                "FEMALEGENITALIACOVERED",
                "MALEGENITALIACOVERED",
                "MALEGENITALIAEXPOSED",
                "MALEBREASTEXPOSED",
                "MALEBREASTCOVERED",
                "FEETCOVERED",
                "FEETEXPOSED",
                "ARMPITSCOVERED",
                "ARMPITSEXPOSED",
                "ANUSCOVERED",
                "ANUSEXPOSED"];
            browser.storage.sync.set({
                labels: labels
            });
        }
        // Censor Type
        if (res.censor_type != null) {
            censor_type = res.censor_type;
        } else {
            censor_type = 'black';
            browser.storage.sync.set({
                censor_type: 'black'
            });
        }
        // File Types
        if (res.file_types != null) {
            file_types = res.file_types;
        } else {
            file_types = ['jpg', 'png', 'bmp', 'webp'];
            browser.storage.sync.set({
                file_types: ['jpg', 'png', 'bmp', 'webp']
            });
        }

        // CensorType Configuration
        if (res.blur_configuration != null) {
            blur_configuration = res.blur_configuration;
        } else {
            browser.storage.sync.set({
                blur_configuration: default_blur_configuration,
            });
        }
        if (res.pixel_configuration != null) {
            pixel_configuration = res.pixel_configuration;
        } else {
            browser.storage.sync.set({
                pixel_configuration: default_pixel_configuration,
            });
        }
        if (res.bar_configuration != null) {
            bar_configuration = res.bar_configuration;
        } else {
            browser.storage.sync.set({
                bar_configuration: default_bar_configuration,
            });
        }

        if (res.glitch_configuration != null) {
            glitch_configuration = res.glitch_configuration;
        } else {
            browser.storage.sync.set({
                glitch_configuration: default_glitch_configuration,
            });
        }

        if (res.gif_configuration != null) {
            gif_configuration = res.gif_configuration;
        } else {
            browser.storage.sync.set({
                gif_configuration: default_gif_configuration,
            });
        }

        if (res.whiteblacklist_configuration != null) {
            whiteblacklist_configuration = res.whiteblacklist_configuration;
        } else {
            browser.storage.sync.set({
                whiteblacklist_configuration: default_whiteblacklist_configuration,
            });
        }

        repairConfigurations();

        refreshLogin();
    });
}

function updateOptions(startup) {
    let options_request = browser.storage.sync.get(['active', 'labels', 'censor_type', 'file_types', 'bar_color', 'debug',
        'do_cache', 'bar_configuration', 'blur_configuration', 'pixel_configuration', 'glitch_configuration', 'gif_configuration','whiteblacklist_configuration']);
    options_request.then((res) => {

        debug = res.debug || false;
        if (extension_active !== res.active || startup) {
            browser.browsingData.remove({},
                {
                    cache: true,
                });
            log("Cache cleared");
        }

        extension_active = res.active;
        labels = res.labels || [];
        censor_type = res.censor_type || 'black';
        file_types = res.file_types || ['jpg', 'png', 'bmp', 'webp'];
        do_cache = res.do_cache;
        blur_configuration = res.blur_configuration || default_blur_configuration;
        bar_configuration = res.bar_configuration || default_bar_configuration;
        //bar_configuration._color = hexConverter(res.bar_color) || hexConverter('#000000');
        pixel_configuration = res.pixel_configuration || default_pixel_configuration;
        glitch_configuration = res.glitch_configuration || default_glitch_configuration;
        gif_configuration = res.gif_configuration || default_gif_configuration;
        whiteblacklist_configuration = res.whiteblacklist_configuration || default_whiteblacklist_configuration;
    });
}

initializeOptions();


/**********************************************************************************************************************/


/**********************************************************************************************************************
 * AI
 **********************************************************************************************************************/
const [modelWeight, modelHeight] = [320, 320];
const weights = browser.runtime.getURL("web_model/model.json");
var model = null;
tf.loadGraphModel(weights).then(model_await => {

    model = model_await;

    //Warmup
    const dummy = tf.zeros([1, 320, 320, 3], 'float32');
    return model.executeAsync(dummy).then(function (result) {
        dummy.dispose();
        log("Model LOADED");
        return result;
    });


});

browser.webRequest.onBeforeRequest.addListener(
    blockImages,
    {urls: ["<all_urls>"], types: ["image", "imageset", "media", "main_frame"]},
    ["blocking"]
);

browser.tabs.onRemoved.addListener(handleRemoved);

function handleRemoved(tabId, removeInfo) {
    log("Tab: " + tabId + " is closing");
    log("Window ID: " + removeInfo.windowId);
    log("Window is closing: " + removeInfo.isWindowClosing);
    log("queue length before: ", queue.length);
    queue.forEach(function (item, index, object) {
        if (item.tabId == tabId) {
            object.splice(index, 1);
            log('Removed from queue', object);
        }
    });
    log("Queue length after: ", queue.length);

}

browser.tabs.onUpdated.addListener(handleUpdated);

function handleUpdated(tabId, changeInfo, tabInfo) {
    log("Updated tab: " + tabId);
    log("Changed attributes: ");
    log(changeInfo);
    log("New tab Info: ");
    log(tabInfo);
    if (tabInfo.active) {
        if (current_tab_id != tabInfo.id) {
            current_tab_id = tabInfo.id;
            log("current_tab_id", current_tab_id);
            queue.sort(compareCensorRequest);
        }
    }

    // What if we open a new web-page but there are ongion requests
    // This is bugged
    /*
    queue.forEach(function (item, index, object) {
      if(item.tabId == tabId && item.originUrl != tabInfo.url){
          object.splice(index, 1);
          log("Removing ",tabId, item.originUrl, tabInfo);
      }
    });
    */

}


browser.webRequest.onHeadersReceived.addListener(
    uncache,
    {urls: ["<all_urls>"], types: ["image", "imageset", "media", "main_frame"]}, //"main_frame"
    ["blocking", "responseHeaders"]
);

function uncache(e) {
    if (!extension_active) {
        return;
    }
    log("UNCACHING:", e);
    var asyncCacheControl = new Promise((resolve, reject) => {
        for (var i = e.responseHeaders.length - 1; i >= 0; --i) {
            if (e.responseHeaders[i].name == "Cache-Control") {
                e.responseHeaders.splice(i, 1);
            }
        }
        var cacheControl = {
            name: "Cache-Control",
            value: "no-cache, no-store, must-revalidate"
        };
        e.responseHeaders.push(cacheControl);
        resolve({responseHeaders: e.responseHeaders});
    });
    return asyncCacheControl;
}


function compareCensorRequest(a, b) {
    /*
    if(a.tabId != current_tab_id){
        return 1;
    }
    */
    if (a.tabId != b.tabId) {
        if (a.tabId == current_tab_id) {
            //log("Sortiere a nach vorne weil current ", a.tabId, "current = ", current_tab_id);
            return -1;
        }
        if (b.tabId == current_tab_id) {
            //log("Sortiere a nach vorne weil nicht current", a.tabId, "current = ", current_tab_id);
            return 1;
        }
    }
    if (a.img_dimension > b.img_dimension) {
        //log("Sortiere größeres Biild nach vorne");
        return -1;
    }
    if (a.img_dimension < b.img_dimension) {
        //log("Sortiere kleines Bild nach hinten");
        return 1;
    }
    return 0;
}

function blockImages(requestDetails) {
    if (!extension_active) {
        return {};
    }

    if(whiteblacklist_configuration.mode > 0){
        if(whiteblacklist_configuration.mode == 1){
            let whitelisted = false;
            for (let regex of whiteblacklist_configuration.white_list) {
               if(new RegExp(regex).test(requestDetails.url)){
                   whitelisted = true;
                   break;
               }
            }
            if(!whitelisted){
                return {};
            }
        }else if(whiteblacklist_configuration.mode == 2){
            for (let regex of whiteblacklist_configuration.black_list) {
                if(new RegExp(regex).test(requestDetails.url)){
                    return {};
                }
            }
        }
    }

    let str = requestDetails.url.substring(0, requestDetails.url.length - 4);
    let pathArray = requestDetails.url.split('/');
    let protocol = pathArray[0];
    let host = pathArray[2];
    let favicon_url = protocol + '//' + host + '/favicon';
    if (favicon_url === str) {
        return {};
    }



    let filter = browser.webRequest.filterResponseData(requestDetails.requestId);

    filter.onerror = event => {
        log(`Error: ${filter.error}`);
    }
    let stream = [];
    let promise;

    let streaming = false;

    filter.onstart = event => {
        promise = new Promise((ai_resolve, ai_reject) => {
            requestContentType(requestDetails.url, function (content_type_header) {
                //When we encounter a video we do not want to wait until the filter has finished
                if (content_type_header != null && content_type_header.startsWith('video')) {
                    streaming = true;
                    ai_reject();
                    return {};
                }else {
                    toDataURL(requestDetails.url, function (imgdata, contenttype) {
                        if (!processableContentType(contenttype, file_types, gif_configuration)) {
                            if (contenttype === 'application/xml') {
                                filter.disconnect();
                                return {};
                            }
                            ai_reject("rejected by contenttype " + contenttype);
                            return {};
                        }
                        let crf = context_requests.filter(e => e.url === requestDetails.url);
                        crf.forEach(f => context_requests.splice(context_requests.findIndex(e => e.name === f.name), 1));

                        if (do_cache && contenttype !== "image/gif") {
                            let cache_url = requestDetails.url;
                            if (crf.length > 0) {
                                cache_url = crf[0].originalUrl;
                            }
                            let cached_result = censored_cache.get(cache_url);
                            let censor_type_local = censor_type;
                            if (crf.length > 0) {
                                censor_type_local = crf[0].censor_type;
                            }
                            if (cached_result) {
                                handleCache(cached_result, requestDetails, filter, censor_type_local);
                                ai_resolve();
                                return {};
                            }
                        }
                        let blob = new Blob([imgdata], {type: contenttype});
                        let filename = requestDetails.url.split('/').pop();
                        let file = new File([blob], filename);
                        let censor_type_processed = censor_type;
                        if (crf.length > 0) {
                            censor_type_processed = crf[0].censor_type;
                        }
                        //context_requests = context_requests.filter(e => e.url !== requestDetails.url);

                        if ((!gif_configuration._thumbnails || crf.length > 0) && contenttype === "image/gif" &&
                            (file_types.includes('gif') || crf.length > 0)
                        ) {
                            if (user != null && user.permissions.permission_file_type_gif <= user.patreon_tier) {
                                if (gif_configuration._memory_method > 0) {
                                    log("processGIFMem (fast)");
                                    processGIFMem(requestDetails, blob, filter, censor_type_processed);
                                } else {
                                    log("processGIF");
                                    processGIF(requestDetails, blob, filter, censor_type_processed);
                                }
                            } else {
                                ai_reject("No permissions for GIF or user object not established");
                            }
                        } else {
                            checkMime(file, function (type, ext, header, cont) {
                                if (!cont && crf.length == 0) {
                                    ai_reject("rejected by mime" + requestDetails.url);
                                    return {};
                                }
                                //let t0 = performance.now();
                                processImage(requestDetails, blob, file, imgdata, filename, contenttype, ext, header, filter, censor_type_processed).then(r => {
                                    //let t1 = performance.now();
                                    //console.log("processImage time " + (t1 - t0) + " millisseconds.");
                                    //console.log("hallo welt")
                                    return r
                                }).catch(e => {
                                    ai_reject("error in processImage");
                                });
                                return {};
                            }, file_types, gif_configuration);
                        }
                    });
                }
         });
        });
        promise.catch((e) => {
            log("Error: ", e);
            promise.error = true;
        });
    }

    filter.ondata = event => {
        if(streaming){
            for (let data of stream) {
                filter.write(data);
            }
            stream = [];
            filter.write(event.data);
        }else{
            stream.push(event.data);
        }

    }

    filter.onstop = event => {
        if(streaming){
            filter.disconnect();
            return;
        }
        if (promise && promise.error) {
            //console.log("Handling error by writing original data stream");
            for (let data of stream) {
                filter.write(data);
            }
            filter.disconnect();
        } else if(promise) {
            promise.catch((e) => {
                //console.log("Handling error by writing original data stream", e);
                for (let data of stream) {
                    filter.write(data);
                }
                filter.disconnect();
            });
        }


    }

}


function handleCache(cached_result, requestDetails, filter, censor_type_local = null) {
    log("This request was loaded from cache: ", cached_result);
    if (
        cached_result.censortype == censor_type &&
        cached_result.labels.length === labels.length &&
        cached_result.labels.every(function (value, index) {
            return value === labels[index]
        })
    ) {
        //This is for future purpose
        cached_result.load_from_cache = true;
    } else {
        cached_result.load_from_cache = false;
    }
    //New options to apply:
    if (censor_type_local == null) {
        cached_result.censortype = censor_type;
    } else {
        cached_result.censortype = censor_type_local;
    }

    cached_result.labels = labels;
    applyFilter(requestDetails, cached_result, filter);
}


function processGIF(requestDetails, blob, filter, censor_type_local = null) {
    //Queue.enqueue(() => new Promise((qresolve, reject) =>
    //{
    if (censor_type_local == null) {
        censor_type_local = censor_type;
    }
    let t0 = performance.now();
    blob.arrayBuffer().then(buffer => {
        let gif = window.gifuct.parseGIF(buffer);
        let frames = window.gifuct.decompressFrames(gif, true);
        let gif_encoder = new GIF({
            workers: 2,
            quality: 10,
            debug: debug,
            //background: '#fff',
            //transparent: 'rgba(0,0,0,255)',
            //transparent: 0,
        });
        let gifpromises = [];
        let last_frame = null;
        let transparency_set = false;
        let gif_base_width = null;
        let gif_base_height = null;
        frames.forEach(function (element, idx, array) {
            if (idx === 0) {
                gif_base_width = element.dims.width;
                gif_base_height = element.dims.height;
                gif_encoder.setOption('width', gif_base_width);
                gif_encoder.setOption('height', gif_base_height);
                log("GIF ENC Initialised");

                /* return a single frame
                var tempCanvas = document.createElement('canvas');
                tempCanvas.width = element.dims.width;
                tempCanvas.height = element.dims.height;
                var tempCtx = tempCanvas.getContext('2d');
                var frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                frameImageData.data.set(element.patch);
                tempCtx.putImageData(frameImageData, 0, 0);
                tempCanvas.toBlob(function(blob) {
                    var reader  = new FileReader();
                    reader.onloadend = function () {
                        resolve({redirectUrl: reader.result});
                    }
                    reader.readAsDataURL(blob);
                });
                return;
                */
            }
            if (element.disposalType > 1 && !transparency_set && 'transparentIndex' in element) {
                gif_encoder.setOption('transparent', 'rgba(0,0,0,${element.transparentIndex})');
                transparency_set = true;
            }

            //log("frames", element);
            let tempCanvas = document.createElement('canvas');
            let tempCtx = tempCanvas.getContext('2d');
            tempCanvas.width = gif_base_width;//element.dims.width;
            tempCanvas.height = gif_base_height;//element.dims.height;
            //tempCanvas.width = element.dims.width;
            //tempCanvas.height = element.dims.height;

            //var naturalHeight = element.dims.height;
            //var naturalWidth = element.dims.width;
            let frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
            //let frameImageData = tempCtx.createImageData(gif_base_width, gif_base_height);

            //let dif_width = Math.max(gif_base_width-element.dims.width, 0);
            //let dif_height = Math.max(gif_base_height-element.dims.height, 0);

            if (element.disposalType <= 1 && 'transparentIndex' in element && last_frame) {
                //frameImageData.data.set(last_frame.patch);
                //frameImageData = last_frame;

                frameImageData.data.set(element.patch);

                tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                let correctedFrameImageData = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);

                frameImageData = mergeImageData([last_frame, correctedFrameImageData]);
                //tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top,0,0, element.dims.width, element.dims.height);
                tempCtx.putImageData(frameImageData, 0, 0);
            } else {
                frameImageData.data.set(element.patch);
                tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
            }
            last_frame = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);
            //const input = tf.image.resizeBilinear(tf.browser.fromPixels(tempCanvas), [modelWeight, modelHeight]).div(255.0).expandDims(0);
            const fromPixels = tf.browser.fromPixels(tempCanvas);
            const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
            const divved = resized.div(255.0);
            const input = divved.expandDims(0);
            fromPixels.dispose();
            resized.dispose();
            divved.dispose();
            let promise = //new Promise(function(resolve, reject) {
                Queue.enqueue(() => new Promise((resolve, reject) => {
                    model.executeAsync(input).then(res => {
                        const [boxes, scores, classes, valid_detections] = res;
                        let valid_detections_sync = valid_detections.dataSync();
                        let classes_sync = classes.dataSync();
                        let boxes_sync = boxes.dataSync();
                        let scores_sync = scores.dataSync();
                        valid_detections.dispose();
                        classes.dispose();
                        scores.dispose();
                        boxes.dispose();
                        let marvinImage = new MarvinImage();
                        tempCanvas.toBlob(function (blob) {
                            let urlCreator = window.URL || window.webkitURL;
                            let marvinUrl = urlCreator.createObjectURL(blob);
                            marvinImage.load(marvinUrl, function () {
                                let processedRequest = new ProcessedCensorRequest(requestDetails.url, blob, gif_base_width, gif_base_height, 'image/gif',
                                    boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                                    censor_type_local, labels);
                                paintCensorRequest(marvinImage, processedRequest, bar_configuration, pixel_configuration, blur_configuration, glitch_configuration);
                                tempCtx.putImageData(marvinImage.imageData, 0, 0);
                                resolve(
                                    {
                                        frame: idx,
                                        delay: element.delay ? element.delay : 100,
                                        disposalType: element.disposalType,
                                        processedData: tempCtx
                                    });

                            });

                        });
                    }).catch(function (err) {
                        log("gif error", err);
                        resolve(
                            {
                                frame: idx,
                                delay: element.delay,
                                disposalType: element.disposalType,
                                processedData: tempCtx
                            }
                        );
                    }).finally(fin => {
                        input.dispose();
                        log('Tensorcount:', tf.memory().numTensors);
                    });

                }), requestDetails.tabId, requestDetails.originUrl, requestDetails.url, blob, gif_base_width, gif_base_height);
            gifpromises.push(promise);
        });

        Promise.allSettled(gifpromises).then((results) => results.forEach(
            (result) => {
                gif_encoder.addFrame(result.value.processedData, {
                    delay: result.value.delay,
                    dispose: result.value.disposalType
                });
            }
        )).finally(function (err) {
            gif_encoder.on('finished', function (blob) {
                blob.arrayBuffer().then(buffer => {
                    let t1 = performance.now();
                    log("Processing GIF time " + (t1 - t0) + " millisseconds.");
                    try {
                        filter.write(buffer);
                        filter.close();
                        //resolve();
                    } catch (err) {
                        log('Caught ' + err, buffer);
                    }
                });
            });
            gif_encoder.render();

        });
    });
    //}), requestDetails.tabId, requestDetails.originUrl, requestDetails.url, blob, 9999
    //);

}

function processGIFMem(requestDetails, blob, filter, censor_type_local = null) {
    Queue.enqueue(() => new Promise((resolve, reject) => {
            if (censor_type_local == null) {
                censor_type_local = censor_type;
            }
            let t0 = performance.now();
            blob.arrayBuffer().then(buffer => {
                let gif = window.gifuct.parseGIF(buffer);
                let frames = window.gifuct.decompressFrames(gif, true);
                let gif_encoder = new GIF({
                    workers: 2,
                    quality: 10,
                    debug: debug,
                    //background: '#fff',
                    //transparent: 'rgba(0,0,0,255)',
                    //transparent: 0,
                });
                let gifpromises = [];
                let last_frame = null;
                let transparency_set = false;
                let gif_base_width = null;
                let gif_base_height = null;
                frames.forEach(function (element, idx, array) {
                    if (idx === 0) {
                        gif_base_width = element.dims.width;
                        gif_base_height = element.dims.height;
                        gif_encoder.setOption('width', gif_base_width);
                        gif_encoder.setOption('height', gif_base_height);
                        log("GIF ENC Initialised");

                        /* return a single frame
                        var tempCanvas = document.createElement('canvas');
                        tempCanvas.width = element.dims.width;
                        tempCanvas.height = element.dims.height;
                        var tempCtx = tempCanvas.getContext('2d');
                        var frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                        frameImageData.data.set(element.patch);
                        tempCtx.putImageData(frameImageData, 0, 0);
                        tempCanvas.toBlob(function(blob) {
                            var reader  = new FileReader();
                            reader.onloadend = function () {
                                resolve({redirectUrl: reader.result});
                            }
                            reader.readAsDataURL(blob);
                        });
                        return;
                        */
                    }
                    if (element.disposalType > 1 && !transparency_set && 'transparentIndex' in element) {
                        gif_encoder.setOption('transparent', 'rgba(0,0,0,${element.transparentIndex})');
                        transparency_set = true;
                    }
                    let promise = new Promise(function (resolve, reject) {
                        //log("frames", element);
                        let tempCanvas = document.createElement('canvas');
                        let tempCtx = tempCanvas.getContext('2d');
                        tempCanvas.width = gif_base_width;//element.dims.width;
                        tempCanvas.height = gif_base_height;//element.dims.height;
                        //tempCanvas.width = element.dims.width;
                        //tempCanvas.height = element.dims.height;

                        //var naturalHeight = element.dims.height;
                        //var naturalWidth = element.dims.width;
                        let frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                        //let frameImageData = tempCtx.createImageData(gif_base_width, gif_base_height);

                        //let dif_width = Math.max(gif_base_width-element.dims.width, 0);
                        //let dif_height = Math.max(gif_base_height-element.dims.height, 0);

                        if (element.disposalType <= 1 && 'transparentIndex' in element && last_frame) {
                            //frameImageData.data.set(last_frame.patch);
                            //frameImageData = last_frame;

                            frameImageData.data.set(element.patch);

                            tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                            let correctedFrameImageData = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);

                            frameImageData = mergeImageData([last_frame, correctedFrameImageData]);
                            //tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top,0,0, element.dims.width, element.dims.height);
                            tempCtx.putImageData(frameImageData, 0, 0);
                        } else {
                            frameImageData.data.set(element.patch);
                            tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                        }
                        last_frame = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);
                        //const input = tf.image.resizeBilinear(tf.browser.fromPixels(tempCanvas), [modelWeight, modelHeight]).div(255.0).expandDims(0);
                        const fromPixels = tf.browser.fromPixels(tempCanvas);
                        const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
                        const divved = resized.div(255.0);
                        const input = divved.expandDims(0);
                        fromPixels.dispose();
                        resized.dispose();
                        divved.dispose();
                        log('Tensorcount:', tf.memory().numTensors);
                        model.executeAsync(input).then(res => {
                            input.dispose();
                            const [boxes, scores, classes, valid_detections] = res;
                            let valid_detections_sync = valid_detections.dataSync();
                            let classes_sync = classes.dataSync();
                            let boxes_sync = boxes.dataSync();
                            let scores_sync = scores.dataSync();
                            valid_detections.dispose();
                            classes.dispose();
                            scores.dispose();
                            boxes.dispose();
                            let marvinImage = new MarvinImage();
                            tempCanvas.toBlob(function (blob) {
                                let urlCreator = window.URL || window.webkitURL;
                                let marvinUrl = urlCreator.createObjectURL(blob);
                                marvinImage.load(marvinUrl, function () {
                                    let processedRequest = new ProcessedCensorRequest(requestDetails.url, blob, gif_base_width, gif_base_height, 'image/gif',
                                        boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                                        censor_type_local, labels);
                                    paintCensorRequest(marvinImage, processedRequest, bar_configuration, pixel_configuration, blur_configuration, glitch_configuration);
                                    tempCtx.putImageData(marvinImage.imageData, 0, 0);
                                    resolve(
                                        {
                                            frame: idx,
                                            delay: element.delay ? element.delay : 100,
                                            disposalType: element.disposalType,
                                            processedData: tempCtx
                                        });
                                });
                            });
                        }).catch(function (err) {
                            log("gif error", err);
                            resolve(
                                {
                                    frame: idx,
                                    delay: element.delay,
                                    disposalType: element.disposalType,
                                    processedData: tempCtx
                                }
                            );
                        });

                    });
                    gifpromises.push(promise);
                });

                Promise.allSettled(gifpromises).then((results) => results.forEach(
                    (result) => {
                        gif_encoder.addFrame(result.value.processedData, {
                            delay: result.value.delay,
                            dispose: result.value.disposalType
                        });
                    }
                )).finally(function (err) {
                    gif_encoder.on('finished', function (blob) {
                        blob.arrayBuffer().then(buffer => {
                            let t1 = performance.now();
                            log("Processing time " + (t1 - t0) + " millisseconds.");
                            try {
                                filter.write(buffer);
                                filter.close();
                                resolve();
                            } catch (err) {
                                log('Caught ' + err, buffer);
                            }
                        });
                    });
                    gif_encoder.render();

                });


            });
        }), requestDetails.tabId, requestDetails.originUrl, requestDetails.url, blob, 9999
    );

}

function mergeImageData(imageDataArray) {

    let newImageData = imageDataArray[0];

    for (let j = 0; j < imageDataArray.length; j++) { // iterate through the imageDataArray

        for (let i = 0, bytes = imageDataArray[j].data.length; i < bytes; i += 4) { // iterate through image bytes

            let index = (imageDataArray[j].data[i + 3] === 0 ? 0 : j);

            newImageData.data[i] = imageDataArray[index].data[i];
            newImageData.data[i + 1] = imageDataArray[index].data[i + 1];
            newImageData.data[i + 2] = imageDataArray[index].data[i + 2];
            newImageData.data[i + 3] = imageDataArray[index].data[i + 3];
        }
    }
    return newImageData;
};

function processImage(requestDetails, blob, file, imgdata, filename, contenttype, ext, header, filter, censor_type_local = null) {
    return new Promise((resolve, reject) => {

        if (!filename.endsWith("." + ext)) {
            filename += "." + ext;
            file = new File([blob], filename);
        }

        if (file.size < 1000) { // = < 10 kb
            log("rejected by filesize", file.size);
            reject("rejected by filesize");
            return;
        }
        if (censor_type_local == null) {
            censor_type_local = censor_type;
        }
        let urlCreator = window.URL || window.webkitURL;
        let imageUrl = urlCreator.createObjectURL(blob);
        let c = document.createElement('img');
        c.src = imageUrl;
        let imageData = new Image();
        imageData.onload = function () {
            let naturalWidth = this.naturalWidth;
            let naturalHeight = this.naturalHeight;
            if (this.naturalWidth < 51 || this.naturalHeight < 51) {
                //filter.disconnect();
                reject("rejected by image dimensions");
                return;
            }
            let img_dimension = this.naturalWidth * this.naturalHeight;
            Queue.enqueue(
                () => new Promise((resolve, reject) => {
                    //const input = tf.image.resizeBilinear(tf.browser.fromPixels(c), [modelWeight, modelHeight]).div(255.0).expandDims(0);
                    const fromPixels = tf.browser.fromPixels(c);
                    const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
                    const divved = resized.div(255.0);
                    const input = divved.expandDims(0);
                    let t0 = performance.now();
                    model.executeAsync(input).then(res => {
                        let t1 = performance.now();
                        //console.log("model.executeAsync time " + (t1 - t0) + " millisseconds.");
                        const [boxes, scores, classes, valid_detections] = res;
                        let valid_detections_sync = valid_detections.dataSync();
                        let classes_sync = classes.dataSync();
                        let boxes_sync = boxes.dataSync();
                        let scores_sync = scores.dataSync();
                        valid_detections.dispose();
                        classes.dispose();
                        scores.dispose();
                        boxes.dispose();
                        log(valid_detections_sync, classes_sync, boxes_sync, scores_sync);
                        var processedRequest = new ProcessedCensorRequest(requestDetails.url, imgdata, naturalWidth, naturalHeight, contenttype,
                            boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                            censor_type_local, labels);
                        if (do_cache && !censored_cache.get(requestDetails.url)) {
                            censored_cache.set(requestDetails.url, processedRequest);
                        }
                        applyFilter(requestDetails, processedRequest, filter)
                            .catch(e => {
                                reject('applyFilter filter error');
                                return;
                            });

                    }).catch(e => {
                        log(e.message)
                        reject(e.message);
                        return;
                    }).finally(fin => {
                        resized.dispose();
                        divved.dispose();
                        input.dispose();
                        fromPixels.dispose();
                        resolve({});
                        return;
                    });
                    log('Tensorcount:', tf.memory().numTensors);
                }).catch(e => {
                    log('Tensorflowerror outer catch', e, requestDetails.url);
                    reject('Tensorflowerror outer catch');
                }).then((val) => resolve(val))
                , requestDetails.tabId, requestDetails.originUrl, requestDetails.url, imageData, img_dimension)
            queue.sort(compareCensorRequest);
        };
        imageData.src = imageUrl;

    });
}

function applyFilter(requestDetails, processedRequest, filter) {
    let promise = new Promise((resolve, reject) => {
        let imgdata = processedRequest.imgdata;
        let contenttype = processedRequest.contenttype;
        let marvinImage = new MarvinImage();
        let urlCreator = window.URL || window.webkitURL;
        let blob = new Blob([imgdata], {type: contenttype});
        let marvinUrl = urlCreator.createObjectURL(blob);
        marvinImage.load(marvinUrl, function () {
            let t0 = performance.now();
            paintCensorRequest(marvinImage, processedRequest, bar_configuration, pixel_configuration, blur_configuration, glitch_configuration);
            let t1 = performance.now();
            //console.log("paintCensorRequest time " + (t1 - t0) + " millisseconds.");
            let marvinBlob = marvinImage.toBlob();
            marvinBlob.arrayBuffer().then(buffer => {
                try {
                    filter.write(buffer);
                    filter.disconnect();
                    resolve(buffer);
                } catch (err) {
                    log('Caught ' + err, buffer);
                    reject();
                }
            });
        });
    });
    return promise;
}
