document.addEventListener('DOMContentLoaded', restoreOptions);

function restoreOptions() {
	restoreONOFF();
	restoreLabels();
	restoreFiletypes();
	restoreCensorOptions();
	restoreCensorConfigs();
	restoreLoginState();
	restoreDebug();
	restoreAdvanced();
	restoreWhiteBlackList();
	updateHelp();
}
function restoreONOFF() {
	let onoff = browser.storage.sync.get('active');
	onoff.then((res) => {
		if(res.active){
			$('.onoff').attr('check', res.active);
		}else{
			$('.onoff').removeAttr('check');
		}
		$('#onoff').html(res.active? 'ON' : 'OFF');
	});
}

function restoreLabels() {
	let labels = browser.storage.sync.get('labels');
	labels.then((res) => {
		$('.labelButton').each(function(i, obj) {
			if(res.labels.includes(obj.id)){
				$(this).addClass("selected");
				$('button[label='+$(this).attr('id')+']').attr('check', true);
				$('button[label='+$(this).attr('id')+']').parent().attr('check', true);
			}else{
				$(this).removeClass("selected");
				$('button[label='+$(this).attr('id')+']').removeAttr('check');
				$('button[label='+$(this).attr('id')+']').parent().removeAttr('check');
			}
		});
	});
}

function restoreFiletypes() {
	let file_types = browser.storage.sync.get('file_types');
	file_types.then((res) => {
		$('.filetypeToogle').each(function(i, obj) {
			if(res.file_types.includes($(this).attr('label'))){
				$(this).attr('check', true);
			}else{
				$(this).removeAttr('check');
			}
		});
	});
}

function restoreCensorOptions() {
	let censor_type = browser.storage.sync.get(['censor_type']);
	censor_type.then((res) => {
		$('.censormodeToogle').each(function(i, obj) {
			if(res.censor_type.includes($(this).attr('label'))){
				$(this).attr('check', true);
				$("#"+$(this).attr('label')).addClass("selected");
			}else{
				$(this).removeAttr('check');
			}
		});
		//$('#bar_colorpicker').attr('value', (res.bar_color));
	});

}

function restoreLoginState(){
	let user_req = browser.storage.sync.get(['user', 'username', 'password']);
	user_req.then((res) => {
		if(res.user){
			loginSuccess(res.user);
		}else{
			if(res.username){
				$("#profile_name").html(res.username);
				$("#profile_tier").html(res.password || '');
			}
			$("#login_form :input").prop("disabled", false);
			$("#login_form").show("slow");
			$("#profile").hide();
		}
	});
}

function restoreDebug() {
	let debug = browser.storage.sync.get('debug');
	debug.then((res) => {
		$('.debugToogle').each(function(i, obj) {
			if(res.debug){
				$(this).attr('check', true);
				$('#'+$(this).attr('label')).addClass("selected");
			}else{
				$(this).removeAttr('check');
				$('#'+$(this).attr('label')).removeClass("selected");
			}

		});
	});
}

function restoreCensorConfigs() {
	let censor_type = browser.storage.sync.get(['bar_configuration','blur_configuration','pixel_configuration','glitch_configuration','gif_configuration']);
	censor_type.then((res) => {
		$('#bar_colorpicker').attr('value', (res.bar_configuration._color));
		$('#bar_scale').val(res.bar_configuration._scale);

		$('#blur_scale').val(res.blur_configuration._scale);
		$('#blur_radius').val(res.blur_configuration._blur_radius);
		$('#blur_strength').val(res.blur_configuration._blur_strength);
		$('#blur_mask_scale').val(res.blur_configuration._mask_scale);

		$('#pixel_geometry').val(res.pixel_configuration._geometry);
		$('#pixel_scale').val(res.pixel_configuration._scale);
		$('#pixel_distance').val(res.pixel_configuration._distance);
		$('#pixel_size').val(res.pixel_configuration._pixel_size);
		$('#pixel_size_mode').val(res.pixel_configuration._pixel_size_mode);

		if($('#pixel_size_mode').val() == 0){
			$('#pixel_size_suffix').val('%');
		}else{
			$('#pixel_size_suffix').val('px');
		}
		$('#glitch_intensity').val(res.glitch_configuration.intensity);
		$('#glitch_scale').val(res.glitch_configuration._scale);
		$('#glitch_horizontal_shift').prop('checked', res.glitch_configuration.shift_horizontal);
		$('#glitch_vertical_shift').prop('checked', res.glitch_configuration.shift_vertical);



		$('#gif_memory_mode').val(res.gif_configuration._memory_method);
		if(res.gif_configuration._thumbnails){
			$('#gif_thumb').attr('check', true);
			$('#gif_thumb_button').attr('check', true);
		}else{
			$('#gif_thumb').removeAttr('check');
			$('#gif_thumb_button').removeAttr('check');
		}
	});
}

function restoreAdvanced() {
	let advanced = browser.storage.sync.get(['video_overlay']);
	advanced.then((res) => {
		if(res.video_overlay){
			$('.video_overlay').attr('check', true);
		}else{
			$('.video_overlay').removeAttr('check');
		}
	});
}

function restoreWhiteBlackList() {
	let whiteblacklist = browser.storage.sync.get(['whiteblacklist_configuration']);
	whiteblacklist.then((res) => {
		if(res.whiteblacklist_configuration.mode == 0){
			$("#radio-nolist").prop("checked", true);
			$('#whitelist-container').hide();
			$('#blacklist-container').hide();
		}else if(res.whiteblacklist_configuration.mode == 1){
			$("#radio-whitelist").prop("checked", true);
			$('#whitelist-container').show();
			$('#blacklist-container').hide();
		}else{
			$("#radio-blacklist").prop("checked", true);
			$('#whitelist-container').hide();
			$('#blacklist-container').show();
		}
		$('#whitelist').val(res.whiteblacklist_configuration.white_list.join('\n'));
		$('#blacklist').val(res.whiteblacklist_configuration.black_list.join('\n'));
	});
}

function updateHelp() {
	let cache = browser.storage.sync.get('do_cache');
	cache.then((res) => {
		$('.do_cacheToogle').each(function(i, obj) {
			if(res.do_cache){
				$(this).attr('check', true);
			}else{
				$(this).removeAttr('check');
			}
		});
	});
	let sending = browser.runtime.sendMessage({
		getdata: true
	});
	sending.then(function(message){
		$('#cached-count').html(message.cache.size);
		let mem = 0;
		message.cache.forEach(function(value, key, map){
			 mem += value._imgdata.byteLength+roughSizeOfObject(value);
		});
		mem = formatBytes(mem);
		$('#cached-size').html(mem);

		/*
		$('#queue-count').html(message.queue.length);
		if(message.queue.length > 0){
			console.log(message.queue[0]);
		}
		*/

		//$('#queue-current').html(message.queue.length > 0?message.queue[0]:'');
	}, handleError);
}

$('#refresh-cache').click(function(){
	updateHelp();
});
$('#clear-cache').click(function(){
	let sending = browser.runtime.sendMessage({
		clearcache: true
	});
	sending.then(function(message){
		updateHelp();
	});
});

$('#pixel_geometry').change(function(){
	if($('#pixel_size_mode').val() == 1){
		if($('#pixel_geometry').val() == 1 || $('#pixel_geometry').val() == 2){
			if($('#pixel_size').val() < 5){
				$('#pixel_size').val(20);
			}
		}
	}
});

$('#pixel_size_mode').change(function(){
	if($('#pixel_size_mode').val() == 1){
		if($('#pixel_geometry').val() == 1 || $('#pixel_geometry').val() == 2){
			if($('#pixel_size').val() < 5){
				$('#pixel_size').val(20);
			}
		}
	}
});


$('.save_config').change(function(){
	let conf = browser.storage.sync.get(['bar_configuration','blur_configuration','pixel_configuration','glitch_configuration','gif_configuration','whiteblacklist_configuration']);
	conf.then((res) => {
		res.bar_configuration._color =  $('#bar_colorpicker').val();
		res.bar_configuration._scale =   Math.max(0.1,parseFloat($('#bar_scale').val()));

		res.blur_configuration._scale =  Math.max(0.1,parseFloat($('#blur_scale').val()));
		res.blur_configuration._mask_scale =  Math.max(0.1,parseFloat($('#blur_mask_scale').val()));
		res.blur_configuration._blur_radius = Math.max(0,parseInt($('#blur_radius').val()));
		res.blur_configuration._blur_strength = Math.max(2,parseInt($('#blur_strength').val()));

		res.pixel_configuration._geometry = $('#pixel_geometry').val();
		res.pixel_configuration._scale = Math.max(0.1,parseFloat($('#pixel_scale').val()));
		res.pixel_configuration._distance = Math.max(0,parseInt($('#pixel_distance').val()));

		res.pixel_configuration._pixel_size = Math.max(1,parseInt($('#pixel_size').val()));
		if($('#pixel_size_mode').val() == 0){
			$('#pixel_size_suffix').html('%');
		}else{
			$('#pixel_size_suffix').html('px');
		}
		res.pixel_configuration._pixel_size_mode = Math.max(0,parseInt($('#pixel_size_mode').val()));

		res.glitch_configuration._scale =  Math.max(0.1,parseFloat($('#glitch_scale').val()));
		res.glitch_configuration.intensity = Math.max(0,parseInt($('#glitch_intensity').val()));
		res.glitch_configuration.shift_horizontal = $('#glitch_horizontal_shift').is(':checked');
		res.glitch_configuration.shift_vertical = $('#glitch_vertical_shift').is(':checked');

		res.gif_configuration._memory_method = Math.max(0,parseInt($('#gif_memory_mode').val()));
		//res.gif_configuration._thumbnails = $('#gif_thumb').val();

		res.gif_configuration._memory_method = Math.max(0,parseInt($('#gif_memory_mode').val()));

		res.whiteblacklist_configuration.mode = $("input:radio[name='wblist']:checked").val();
		let white_list = [];
		$.each($('#whitelist').val().split(/\n/), function(i, line){
			if(line){
				white_list.push(line);
			} else {
				//white_list.push("");
			}
		});
		res.whiteblacklist_configuration.white_list = white_list;

		let black_list = [];
		$.each($('#blacklist').val().split(/\n/), function(i, line){
			if(line){
				black_list.push(line);
			} else {
				//black_list.push("");
			}
		});
		res.whiteblacklist_configuration.black_list = black_list;
		browser.storage.sync.set({
			bar_configuration: res.bar_configuration,
			blur_configuration: res.blur_configuration,
			pixel_configuration: res.pixel_configuration,
			glitch_configuration: res.glitch_configuration,
			gif_configuration: res.gif_configuration,
			whiteblacklist_configuration: res.whiteblacklist_configuration,

		}).then(function(){
			updateOptionsContentScript();
		});

	});
});

$('button[toogle]').click(function(){
	$('#'+$(this).attr('toogle')).toggle();
	$('.'+$(this).attr('toogle')).toggle();
});

$('#radio-nolist[toogle]').click(function(){
	if($(this).is(':checked')){
		$('#whitelist-container').hide();
		$('#blacklist-container').hide();
	}
});
$('#radio-whitelist[toogle]').click(function(){
	if($(this).is(':checked')){
		$('#whitelist-container').show();
		$('#blacklist-container').hide();
	}
});
$('#radio-blacklist[toogle]').click(function(){
	if($(this).is(':checked')){
		$('#whitelist-container').hide();
		$('#blacklist-container').show();
	}
});

$('.tab').click(function(){
	$('.tab').removeClass('tab-selected');
	$(this).addClass('tab-selected');
	$('.content').removeClass('content-displayed');
	$('#'+$(this).attr('content')).addClass('content-displayed');
});

$('io-toggle').click(function(e){
	$(this).toggleAttr('check', true);
	$(this).children().toggleAttr('check', true);
});

$('io-toggle button').click(function(e){
	$(this).toggleAttr('check', true);
	$(this).parent().toggleAttr('check', true);
});

$('.onoff').click(function(e){
	e.stopImmediatePropagation();
	$('#onoff').html($(this).attr('check')? 'ON' : 'OFF');
	activatePuryFi($(this).attr('check')? true : false);
});

$('.labelToogle').click(function(e){
	e.stopImmediatePropagation();
	$('#'+$(this).attr('label')).toggleClass("selected", $(this).attr('check'));
	let labels = browser.storage.sync.get('labels');
	labels.then((res) => {
		let labellist = res.labels || [];
		if($(this).attr('check')){
			if(labellist.indexOf($(this).attr('label')) === -1){
				labellist.push($(this).attr('label'))
			}
		}else{
			if(labellist.indexOf($(this).attr('label')) != -1){
				labellist = removeItemAll(labellist, $(this).attr('label'));
			}
		}
		browser.storage.sync.set({
			labels: labellist
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

$('.overlayToogle').click(function(e){
	e.stopImmediatePropagation();
	browser.storage.sync.set({
		video_overlay: $(this).attr('check')
	}).then(function(){
		updateOptionsContentScript();
	});
});

$('.filetypeToogle').click(function(e){
	e.stopImmediatePropagation();
	let file_types = browser.storage.sync.get('file_types');
	file_types.then((res) => {
		let file_typeslist = res.file_types || [];
		if($(this).attr('check')){
			if(file_typeslist.indexOf($(this).attr('label')) === -1){
				file_typeslist.push($(this).attr('label'))
			}
		}else{
			if(file_typeslist.indexOf($(this).attr('label')) != -1){
				file_typeslist = removeItemAll(file_typeslist, $(this).attr('label'));
			}
		}
		if($('#gif-io').attr('check') && $("#gif_thumb").attr('check')){
			$('#gif_thumb').removeAttr('check');
			$('#gif_thumb_button').removeAttr('check');
			$('.configToogle').click();
		}
		browser.storage.sync.set({
			file_types: file_typeslist
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

$('.censormodeToogle').click(function(e){
	e.stopImmediatePropagation();
	$('button[label=normal]').removeAttr('check');
	$('button[label=normal]').parent().removeAttr('check');
	$('button[label=box]').removeAttr('check');
	$('button[label=box]').parent().removeAttr('check');
	$('button[label=black]').removeAttr('check');
	$('button[label=black]').parent().removeAttr('check');
	$('button[label=pixel]').removeAttr('check');
	$('button[label=pixel]').parent().removeAttr('check');
	$('button[label=blur]').removeAttr('check');
	$('button[label=blur]').parent().removeAttr('check');
	$('button[label=glitch]').removeAttr('check');
	$('button[label=glitch]').parent().removeAttr('check');
	document.getElementById("normal").classList.remove("selected");
	document.getElementById("box").classList.remove("selected");
	document.getElementById("black").classList.remove("selected");
	document.getElementById("pixel").classList.remove("selected");
	document.getElementById("blur").classList.remove("selected");
	document.getElementById("glitch").classList.remove("selected");
	$("#"+$(this).attr('label')).addClass("selected");
	$('button[label='+$(this).attr('label')+']').attr('check', true);
	$('button[label='+$(this).attr('label')+']').parent().attr('check', true);
	let censor_type = browser.storage.sync.get('censor_type');
	censor_type.then((res) => {
		let censor_type = res.censor_type || 'black';
		if($(this).attr('check')){
			censor_type = $(this).attr('label');
		}
		browser.storage.sync.set({
			censor_type: censor_type
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});


$('.debugToogle').click(function(e){
	e.stopImmediatePropagation();
	$('#'+$(this).attr('label')).toggleClass("selected", $(this).attr('check'));
	let debug = browser.storage.sync.get('debug');
	debug.then((res) => {
		let debugging = res.debug || false;
		browser.storage.sync.set({
			debug: ($(this).attr('check')? true : false)
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

$('.configToogle').click(function(e){
	e.stopImmediatePropagation();
	$('#'+$(this).attr('label')).toggleClass("selected", $(this).attr('check'));
	let conf = browser.storage.sync.get(['gif_configuration']);
	conf.then((res) => {
		res.gif_configuration._thumbnails = $('#gif_thumb').attr('check');
		if(res.gif_configuration._thumbnails){
			$('#gif-io').removeAttr('check');
			$('#gif-button').removeAttr('check');
			$('.filetypeToogle').click();
		}
		browser.storage.sync.set({
			gif_configuration: res.gif_configuration,
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

$('.do_cacheToogle').click(function(e){
	e.stopImmediatePropagation();
	$('#'+$(this).attr('label')).toggleClass("selected", $(this).attr('check'));
	let censor = browser.storage.sync.get('do_cache');
	censor.then((res) => {
		browser.storage.sync.set({
			do_cache:  ($(this).attr('check')? true : false)
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

// jquery toggle whole attribute
$.fn.toggleAttr = function(attr, val) {
	var test = $(this).attr(attr);
	if ( test ) {
		// if attrib exists with ANY value, still remove it
		$(this).removeAttr(attr);
	} else {
		$(this).attr(attr, val);
	}
	return this;
};

// jquery toggle just the attribute value
$.fn.toggleAttrVal = function(attr, val1, val2) {
	var test = $(this).attr(attr);
	if ( test === val1) {
		$(this).attr(attr, val2);
		return this;
	}
	if ( test === val2) {
		$(this).attr(attr, val1);
		return this;
	}
	// default to val1 if neither
	$(this).attr(attr, val1);
	return this;
};

$(".labelButton").on('click', function(event){
	$(this).toggleClass("selected");
	if($(this).hasClass("selected")){
		$('button[label='+$(this).attr('id')+']').attr('check', true);
		$('button[label='+$(this).attr('id')+']').parent().attr('check', true);
	}else{
		$('button[label='+$(this).attr('id')+']').removeAttr('check');
		$('button[label='+$(this).attr('id')+']').parent().removeAttr('check');
	}
	let labels = browser.storage.sync.get('labels');
	labels.then((res) => {
		let labellist = res.labels || [];
		if($(this).hasClass("selected")){
			if(labellist.indexOf(this.id) === -1){
				labellist.push(this.id)
			}
		}else{
			if(labellist.indexOf(this.id) != -1){
				labellist = removeItemAll(labellist, this.id);
			}
		}
		browser.storage.sync.set({
			labels: labellist
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

$(".typeButton").on('click', function(event){
	if(!$(this).hasClass("selected")){
		$('button[label=normal]').removeAttr('check');
		$('button[label=normal]').parent().removeAttr('check');
		$('button[label=box]').removeAttr('check');
		$('button[label=box]').parent().removeAttr('check');
		$('button[label=black]').removeAttr('check');
		$('button[label=black]').parent().removeAttr('check');
		$('button[label=pixel]').removeAttr('check');
		$('button[label=pixel]').parent().removeAttr('check');
		$('button[label=blur]').removeAttr('check');
		$('button[label=blur]').parent().removeAttr('check');
		$('button[label=glitch]').removeAttr('check');
		$('button[label=glitch]').parent().removeAttr('check');
		document.getElementById("normal").classList.remove("selected");
		document.getElementById("box").classList.remove("selected");
		document.getElementById("black").classList.remove("selected");
		document.getElementById("pixel").classList.remove("selected");
		document.getElementById("blur").classList.remove("selected");
		document.getElementById("glitch").classList.remove("selected");
		$(this).addClass("selected");

		$('button[label='+$(this).attr('id')+']').attr('check', true);
		$('button[label='+$(this).attr('id')+']').parent().attr('check', true);
		let censor_type = browser.storage.sync.get('censor_type');
		censor_type.then((res) => {
			censor_type = $(this).attr('id');
			browser.storage.sync.set({
				censor_type: censor_type
			}).then(function(){
				updateOptionsContentScript();
			});
		});

	}
});

$(".debugButton").on('click', function(event){
	if(!$(this).hasClass("selected")){
		$(this).addClass("selected");
		$('.debugToogle').attr('check', true);
	}else{
		$(this).removeClass("selected");
		$('.debugToogle').removeAttr('check');
	}
	let debug = browser.storage.sync.get('debug');
	debug.then((res) => {
		let debugging = res.debug || false;
		browser.storage.sync.set({
			debug: $(this).hasClass("selected")
		}).then(function(){
			updateOptionsContentScript();
		});
	});
});

/**
 *	HELPER FUNCTIONS
 */
function removeItemAll(arr, value) {
	let i = 0;
	while (i < arr.length) {
		if (arr[i] === value) {
			arr.splice(i, 1);
		} else {
			++i;
		}
	}
	return arr;
}



